from __future__ import annotations

from pathlib import Path
from typing import TypeVar

import toml
from pydantic import BaseModel, ValidationError

from ..core import Adapter
from ..exceptions import ParseError
from ..exceptions import ValidationError as AdapterValidationError

T = TypeVar("T", bound=BaseModel)


def _ensure_list(d):
    if isinstance(d, list):
        return d
    if isinstance(d, dict) and len(d) == 1 and isinstance(next(iter(d.values())), list):
        return next(iter(d.values()))
    return [d]


class TomlAdapter(Adapter[T]):
    obj_key = "toml"

    @classmethod
    def from_obj(cls, subj_cls: type[T], obj: str | Path, /, *, many=False, **kw):
        try:
            # Handle file path
            if isinstance(obj, Path):
                try:
                    text = Path(obj).read_text()
                except Exception as e:
                    raise ParseError(f"Failed to read TOML file: {e}", source=str(obj))
            else:
                text = obj

            # Check for empty input
            if not text or (isinstance(text, str) and not text.strip()):
                raise ParseError(
                    "Empty TOML content",
                    source=str(obj)[:100] if isinstance(obj, str) else str(obj),
                )

            # Parse TOML
            try:
                parsed = toml.loads(text, **kw)
            except toml.TomlDecodeError as e:
                raise ParseError(
                    f"Invalid TOML: {e}",
                    source=str(text)[:100] if isinstance(text, str) else str(text),
                )

            # Validate against model
            try:
                if many:
                    return [subj_cls.model_validate(x) for x in _ensure_list(parsed)]
                return subj_cls.model_validate(parsed)
            except ValidationError as e:
                raise AdapterValidationError(
                    f"Validation error: {e}",
                    data=parsed,
                    errors=e.errors(),
                )

        except (ParseError, AdapterValidationError):
            # Re-raise our custom exceptions
            raise
        except Exception as e:
            # Wrap other exceptions
            raise ParseError(
                f"Unexpected error parsing TOML: {e}",
                source=str(obj)[:100] if isinstance(obj, str) else str(obj),
            )

    @classmethod
    def to_obj(cls, subj: T | list[T], /, *, many=False, **kw) -> str:
        try:
            items = subj if isinstance(subj, list) else [subj]

            if not items:
                return ""

            payload = (
                {"items": [i.model_dump() for i in items]}
                if many
                else items[0].model_dump()
            )
            return toml.dumps(payload, **kw)

        except Exception as e:
            # Wrap exceptions
            raise ParseError(f"Error generating TOML: {e}")



"""
DataFrame & Series adapters (require `pandas`).
"""

from __future__ import annotations

from typing import TypeVar

import pandas as pd
from pydantic import BaseModel

from ..core import Adapter

T = TypeVar("T", bound=BaseModel)


class DataFrameAdapter(Adapter[T]):
    obj_key = "pd.DataFrame"

    @classmethod
    def from_obj(cls, subj_cls: type[T], obj: pd.DataFrame, /, *, many=True, **kw):
        if many:
            return [subj_cls.model_validate(r) for r in obj.to_dict(orient="records")]
        return subj_cls.model_validate(obj.iloc[0].to_dict(), **kw)

    @classmethod
    def to_obj(cls, subj: T | list[T], /, *, many=True, **kw) -> pd.DataFrame:
        items = subj if isinstance(subj, list) else [subj]
        return pd.DataFrame([i.model_dump() for i in items], **kw)


class SeriesAdapter(Adapter[T]):
    obj_key = "pd.Series"

    @classmethod
    def from_obj(cls, subj_cls: type[T], obj: pd.Series, /, *, many=False, **kw):
        if many:
            raise ValueError("SeriesAdapter supports single records only.")
        return subj_cls.model_validate(obj.to_dict(), **kw)

    @classmethod
    def to_obj(cls, subj: T | list[T], /, *, many=False, **kw) -> pd.Series:
        if many or isinstance(subj, list):
            raise ValueError("SeriesAdapter supports single records only.")
        return pd.Series(subj.model_dump(), **kw)



# Copyright (c) 2025, HaiyangLi <quantocean.li at gmail dot com>
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
from asyncio.log import logger
from collections.abc import Callable
from typing import Any

from pydantic import Field, PrivateAttr

from .temporal import Temporal
from .types import Execution, ExecutionStatus
from .utils import as_async_fn, validate_model_to_dict


class Invokable(Temporal):
    """An executable can be invoked with a request"""

    request: dict | None = None
    execution: Execution = Field(default_factory=Execution)
    response_obj: Any = Field(None, exclude=True)
    _invoke_function: Callable | None = PrivateAttr(None)
    _invoke_args: list[Any] = PrivateAttr([])
    _invoke_kwargs: dict[str, Any] = PrivateAttr({})

    @property
    def has_invoked(self) -> bool:
        return self.execution.status in [
            ExecutionStatus.COMPLETED,
            ExecutionStatus.FAILED,
        ]

    async def _invoke(self):
        if self._invoke_function is None:
            raise ValueError("Event invoke function is not set.")
        async_fn = as_async_fn(self._invoke_function)
        return await async_fn(*self._invoke_args, **self._invoke_kwargs)

    async def invoke(self) -> None:
        start = asyncio.get_event_loop().time()
        response = None
        e1 = None

        try:
            # Use the endpoint as a context manager
            response = await self._invoke()

        except asyncio.CancelledError as ce:
            e1 = ce
            logger.warning("invoke() canceled by external request.")
            raise
        except Exception as ex:
            e1 = ex  # type: ignore

        finally:
            self.execution.duration = asyncio.get_event_loop().time() - start
            if response is None and e1 is not None:
                self.execution.error = str(e1)
                self.execution.status = ExecutionStatus.FAILED
                logger.error(f"invoke() failed for event {str(self.id)[:6]}...")
            else:
                self.response_obj = response
                self.execution.response = validate_model_to_dict(response)
                self.execution.status = ExecutionStatus.COMPLETED
            self.update_timestamp()



"""
AsyncPostgresAdapter - presets AsyncSQLAdapter for PostgreSQL/pgvector.
"""

from __future__ import annotations

from typing import TypeVar

from pydantic import BaseModel

from ..exceptions import ConnectionError
from .async_sql_ import AsyncSQLAdapter

T = TypeVar("T", bound=BaseModel)


class AsyncPostgresAdapter(AsyncSQLAdapter[T]):  # type: ignore[type-arg]
    obj_key = "async_pg"
    DEFAULT = "postgresql+asyncpg://test:test@localhost/test"

    @classmethod
    async def from_obj(cls, subj_cls, obj: dict, /, **kw):
        try:
            # Use the provided DSN if available, otherwise use the default
            engine_url = kw.get("dsn", cls.DEFAULT)
            if "dsn" in kw:
                # Convert the PostgreSQL URL to SQLAlchemy format
                if not engine_url.startswith("postgresql+asyncpg://"):
                    engine_url = engine_url.replace(
                        "postgresql://", "postgresql+asyncpg://"
                    )
            obj.setdefault("engine_url", engine_url)

            # Add PostgreSQL-specific error handling
            try:
                return await super().from_obj(subj_cls, obj, **kw)
            except Exception as e:
                # Check for common PostgreSQL-specific errors
                error_str = str(e).lower()
                if "authentication" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL authentication failed: {e}",
                        adapter="async_pg",
                        url=engine_url,
                    ) from e
                elif "connection" in error_str and "refused" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL connection refused: {e}",
                        adapter="async_pg",
                        url=engine_url,
                    ) from e
                elif "does not exist" in error_str and "database" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL database does not exist: {e}",
                        adapter="async_pg",
                        url=engine_url,
                    ) from e
                # Re-raise the original exception
                raise

        except ConnectionError:
            # Re-raise ConnectionError
            raise
        except Exception as e:
            # Wrap other exceptions
            raise ConnectionError(
                f"Unexpected error in async PostgreSQL adapter: {e}",
                adapter="async_pg",
                url=obj.get("engine_url", cls.DEFAULT),
            ) from e

    @classmethod
    async def to_obj(cls, subj, /, **kw):
        try:
            # Use the provided DSN if available, otherwise use the default
            engine_url = kw.get("dsn", cls.DEFAULT)
            if "dsn" in kw:
                # Convert the PostgreSQL URL to SQLAlchemy format
                if not engine_url.startswith("postgresql+asyncpg://"):
                    engine_url = engine_url.replace(
                        "postgresql://", "postgresql+asyncpg://"
                    )
            kw.setdefault("engine_url", engine_url)

            # Add PostgreSQL-specific error handling
            try:
                return await super().to_obj(subj, **kw)
            except Exception as e:
                # Check for common PostgreSQL-specific errors
                error_str = str(e).lower()
                if "authentication" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL authentication failed: {e}",
                        adapter="async_pg",
                        url=engine_url,
                    ) from e
                elif "connection" in error_str and "refused" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL connection refused: {e}",
                        adapter="async_pg",
                        url=engine_url,
                    ) from e
                elif "does not exist" in error_str and "database" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL database does not exist: {e}",
                        adapter="async_pg",
                        url=engine_url,
                    ) from e
                # Re-raise the original exception
                raise

        except ConnectionError:
            # Re-raise ConnectionError
            raise
        except Exception as e:
            # Wrap other exceptions
            raise ConnectionError(
                f"Unexpected error in async PostgreSQL adapter: {e}",
                adapter="async_pg",
                url=kw.get("engine_url", cls.DEFAULT),
            ) from e



# Copyright (c) 2025, HaiyangLi <quantocean.li at gmail dot com>
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import contextlib
import functools
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from functools import cache
from typing import Any, Protocol, TypeVar
from uuid import UUID

from pydantic import BaseModel

Imp = TypeVar("Imp")


class HasLen(Protocol):
    def __len__(self) -> int: ...


Bin = list[int]
T = TypeVar("T")

__all__ = (
    "get_bins",
    "import_module",
    "sha256_of_dict",
    "convert_to_datetime",
    "validate_uuid",
    "validate_model_to_dict",
    "is_package_installed",
    "is_coroutine_function",
    "as_async_fn",
)


def import_module(
    package_name: str,
    module_name: str | None = None,
    import_name: str | list | None = None,
) -> Imp | list[Imp] | Any:
    """Import a module by its path."""
    try:
        full_import_path = (
            f"{package_name}.{module_name}" if module_name else package_name
        )

        if import_name:
            import_name = (
                [import_name] if not isinstance(import_name, list) else import_name
            )
            a = __import__(
                full_import_path,
                fromlist=import_name,
            )
            if len(import_name) == 1:
                return getattr(a, import_name[0])
            return [getattr(a, name) for name in import_name]
        return __import__(full_import_path)

    except ImportError as e:
        error_msg = f"Failed to import module {full_import_path}: {e}"
        raise ImportError(error_msg) from e


def is_package_installed(package_name: str):
    from importlib.util import find_spec

    return find_spec(package_name) is not None


def get_bins(input_: list[HasLen], /, upper: int) -> list[Bin]:
    """Organizes indices of items into bins based on a cumulative upper limit length.

    Args:
        input_ (list[str]): The list of strings to be binned.
        upper (int): The cumulative length upper limit for each bin.

    Returns:
        list[list[int]]: A list of bins, each bin is a list of indices from the input list.
    """
    current = 0
    bins = []
    current_bin = []
    for idx, item in enumerate(input_):
        if current + len(item) < upper:
            current_bin.append(idx)
            current += len(item)
        else:
            bins.append(current_bin)
            current_bin = [idx]
            current = len(item)
    if current_bin:
        bins.append(current_bin)
    return bins


def sha256_of_dict(obj: dict) -> str:
    """Deterministic SHA-256 of an arbitrary mapping."""
    import hashlib

    import orjson

    payload: bytes = orjson.dumps(
        obj,
        option=(
            orjson.OPT_SORT_KEYS  # canonical ordering
            | orjson.OPT_NON_STR_KEYS  # allow int / enum keys if you need them
        ),
    )
    return hashlib.sha256(memoryview(payload)).hexdigest()


def convert_to_datetime(v):
    if isinstance(v, datetime):
        return v
    if isinstance(v, str):
        with contextlib.suppress(ValueError):
            return datetime.fromisoformat(v)

    error_msg = "Input value for field <created_at> should be a `datetime.datetime` object or `isoformat` string"
    raise ValueError(error_msg)


def validate_uuid(v: str | UUID) -> UUID:
    if isinstance(v, UUID):
        return v
    try:
        return UUID(str(v))
    except Exception as e:
        error_msg = "Input value for field <id> should be a `uuid.UUID` object or a valid `uuid` representation"
        raise ValueError(error_msg) from e


def validate_model_to_dict(v):
    """Serialize a Pydantic model to a dictionary. kwargs are passed to model_dump."""

    if isinstance(v, BaseModel):
        return v.model_dump()
    if v is None:
        return {}
    if isinstance(v, dict):
        return v

    error_msg = "Input value for field <model> should be a `pydantic.BaseModel` object or a `dict`"
    raise ValueError(error_msg)


@cache
def is_coroutine_function(fn, /) -> bool:
    """Check if a function is a coroutine function."""
    return asyncio.iscoroutinefunction(fn)


def force_async(fn: Callable[..., T], /) -> Callable[..., Callable[..., T]]:
    """force a function to be async."""
    pool = ThreadPoolExecutor()

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        future = pool.submit(fn, *args, **kwargs)
        return asyncio.wrap_future(future)  # Make it awaitable

    return wrapper


@cache
def as_async_fn(fn, /):
    """forcefully get the async call of a function"""
    if is_coroutine_function(fn):
        return fn
    return force_async(fn)



import json
from collections.abc import Callable
from functools import wraps
from typing import Any

from ..async_core import AsyncAdapter
from .embedable import Embedable
from .invokable import Invokable
from .types import Embedding, Log
from .utils import as_async_fn, validate_model_to_dict


class Event(Invokable, Embedable):
    event_type: str = None

    def __init__(
        self,
        event_invoke_function: Callable,
        event_invoke_args: list[Any],
        event_invoke_kwargs: dict[str, Any],
        event_type: str = None,
    ):
        super().__init__(response_obj=None)
        self._invoke_function = event_invoke_function
        self._invoke_args = event_invoke_args or []
        self._invoke_kwargs = event_invoke_kwargs or {}
        if event_type is not None:
            self.event_type = event_type

    def create_content(self):
        if self.content is not None:
            return self.content

        event = {"request": self.request, "response": self.execution.response}
        self.content = json.dumps(event, default=str, ensure_ascii=False)
        return self.content

    def to_log(self, event_type: str | None = None, hash_content: bool = False) -> Log:
        if self.content is None:
            self.create_content()

        # Create a Log object with keyword arguments
        log = Log(
            id=str(self.id),
            created_at=self.created_at.isoformat(),
            updated_at=self.updated_at.isoformat(),
            event_type=event_type or self.__class__.__name__,
            content=self.content,
            embedding=self.embedding,
            duration=self.execution.duration,
            status=self.execution.status.value.lower(),  # Ensure status is lowercase
            error=self.execution.error,
        )

        if hash_content:
            from .utils import sha256_of_dict

            log.sha256 = sha256_of_dict({"content": self.content})

        return log


def as_event(
    *,
    request_arg: str | None = None,
    embed_content: bool = False,
    embed_function: Callable[..., Embedding] | None = None,
    adapt: bool = False,
    adapter: type[AsyncAdapter] | None = None,
    event_type: str | None = None,
    **kw,
):
    def decorator(func: Callable):
        @wraps(func)
        async def wrapper(*args, **kwargs) -> Event:
            request_obj = kwargs.get(request_arg) if request_arg else None
            if len(args) > 2 and hasattr(args[0], "__class__"):
                args = args[1:]
            request_obj = args[0] if request_obj is None else request_obj
            event = Event(func, list(args), kwargs, event_type=event_type)
            event.request = validate_model_to_dict(request_obj)
            await event.invoke()
            if embed_content and embed_function is not None:
                async_embed = as_async_fn(embed_function)
                event.embedding = await async_embed(event.content)

            if adapt and adapter is not None:
                await adapter.to_obj(event.to_log(event_type=event_type), **kw)

            return event

        return wrapper

    return decorator



"""
Generic async SQL adapter - SQLAlchemy 2.x asyncio + asyncpg driver.
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import TypeVar

import sqlalchemy as sa
import sqlalchemy.exc as sa_exc
from pydantic import BaseModel, ValidationError
from sqlalchemy.ext.asyncio import create_async_engine

from ..async_core import AsyncAdapter
from ..exceptions import AdapterError, ConnectionError, QueryError, ResourceError
from ..exceptions import ValidationError as AdapterValidationError

T = TypeVar("T", bound=BaseModel)


class AsyncSQLAdapter(AsyncAdapter[T]):
    obj_key = "async_sql"

    # helpers
    @staticmethod
    def _table(meta: sa.MetaData, name: str) -> sa.Table:
        try:
            # In SQLAlchemy 2.x, we should use the connection directly
            return sa.Table(name, meta, autoload_with=meta.bind)
        except sa_exc.NoSuchTableError as e:
            raise ResourceError(f"Table '{name}' not found", resource=name) from e
        except Exception as e:
            raise ResourceError(
                f"Error accessing table '{name}': {e}", resource=name
            ) from e

    # incoming
    @classmethod
    async def from_obj(cls, subj_cls: type[T], obj: dict, /, *, many=True, **kw):
        try:
            # Validate required parameters
            if "engine_url" not in obj:
                raise AdapterValidationError(
                    "Missing required parameter 'engine_url'", data=obj
                )
            if "table" not in obj:
                raise AdapterValidationError(
                    "Missing required parameter 'table'", data=obj
                )

            # Create engine
            try:
                eng = create_async_engine(obj["engine_url"], future=True)
            except Exception as e:
                raise ConnectionError(
                    f"Failed to create async database engine: {e}",
                    adapter="async_sql",
                    url=obj["engine_url"],
                ) from e

            # Execute query
            try:
                # Use a try-except block to handle both real and mocked engines
                try:
                    async with eng.begin() as conn:
                        meta = sa.MetaData()
                        meta.bind = conn
                        tbl = cls._table(meta, obj["table"])
                        stmt = sa.select(tbl).filter_by(**obj.get("selectors", {}))
                        rows = (await conn.execute(stmt)).fetchall()
                except TypeError:
                    # Handle case where eng.begin() is a coroutine in tests
                    if hasattr(eng.begin, "__self__") and hasattr(
                        eng.begin.__self__, "__aenter__"
                    ):
                        # This is for test mocks
                        conn = await eng.begin().__aenter__()
                        meta = sa.MetaData()
                        meta.bind = conn
                        tbl = cls._table(meta, obj["table"])
                        stmt = sa.select(tbl).filter_by(**obj.get("selectors", {}))
                        rows = (await conn.execute(stmt)).fetchall()
                    else:
                        raise
            except ResourceError:
                # Re-raise ResourceError from _table
                raise
            except sa_exc.SQLAlchemyError as e:
                raise QueryError(
                    f"Error executing async SQL query: {e}",
                    query=str(obj.get("selectors", {})),
                    adapter="async_sql",
                ) from e
            except Exception as e:
                raise QueryError(
                    f"Unexpected error in async SQL query: {e}",
                    adapter="async_sql",
                ) from e

            # Handle empty result set
            if not rows:
                if many:
                    return []
                raise ResourceError(
                    "No rows found matching the query",
                    resource=obj["table"],
                    selectors=obj.get("selectors", {}),
                )

            # Convert rows to model instances
            try:
                records = [dict(r) for r in rows]
                if many:
                    return [subj_cls.model_validate(r) for r in records]
                return subj_cls.model_validate(records[0])
            except ValidationError as e:
                raise AdapterValidationError(
                    f"Validation error: {e}",
                    data=records[0] if not many else records,
                    errors=e.errors(),
                ) from e

        except AdapterError:
            raise

        except Exception as e:
            raise QueryError(
                f"Unexpected error in async SQL adapter: {e}", adapter="async_sql"
            )

    # outgoing
    @classmethod
    async def to_obj(
        cls,
        subj: T | Sequence[T],
        /,
        *,
        engine_url: str,
        table: str,
        many=True,
        **kw,
    ):
        try:
            # Validate required parameters
            if not engine_url:
                raise AdapterValidationError("Missing required parameter 'engine_url'")
            if not table:
                raise AdapterValidationError("Missing required parameter 'table'")

            # Create engine
            try:
                eng = create_async_engine(engine_url, future=True)
            except Exception as e:
                raise ConnectionError(
                    f"Failed to create async database engine: {e}",
                    adapter="async_sql",
                    url=engine_url,
                ) from e

            # Prepare data
            items = subj if isinstance(subj, Sequence) else [subj]
            if not items:
                return None  # Nothing to insert

            rows = [i.model_dump() for i in items]

            # Execute insert
            try:
                # Use a try-except block to handle both real and mocked engines
                try:
                    async with eng.begin() as conn:
                        meta = sa.MetaData()
                        meta.bind = conn
                        tbl = cls._table(meta, table)
                        await conn.execute(sa.insert(tbl), rows)
                        return {"inserted_count": len(rows)}
                except TypeError:
                    # Handle case where eng.begin() is a coroutine in tests
                    if hasattr(eng.begin, "__self__") and hasattr(
                        eng.begin.__self__, "__aenter__"
                    ):
                        # This is for test mocks
                        conn = await eng.begin().__aenter__()
                        meta = sa.MetaData()
                        meta.bind = conn
                        tbl = cls._table(meta, table)
                        await conn.execute(sa.insert(tbl), rows)
                        return {"inserted_count": len(rows)}
                    else:
                        raise
            except ResourceError:
                raise

            except sa_exc.SQLAlchemyError as e:
                raise QueryError(
                    f"Error executing async SQL insert: {e}",
                    query=f"INSERT INTO {table}",
                    adapter="async_sql",
                ) from e
            except Exception as e:
                raise QueryError(
                    f"Unexpected error in async SQL insert: {e}",
                    adapter="async_sql",
                ) from e

        except AdapterError:
            raise

        except Exception as e:
            raise QueryError(
                f"Unexpected error in async SQL adapter: {e}", adapter="async_sql"
            )



"""
pydapter - tiny trait + adapter toolkit.
"""

from .async_core import AsyncAdaptable, AsyncAdapter, AsyncAdapterRegistry
from .core import Adaptable, Adapter, AdapterRegistry

__all__ = [
    "Adaptable",
    "Adapter",
    "AdapterRegistry",
    "AsyncAdaptable",
    "AsyncAdapter",
    "AsyncAdapterRegistry",
]
__version__ = "0.1.7"



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # Type checking imports
    from typing import runtime_checkable

    from typing_extensions import Protocol

    from .embedable import Embedable
    from .event import Event
    from .identifiable import Identifiable
    from .invokable import Invokable
    from .temporal import Temporal
    from .types import Embedding, Execution, ExecutionStatus, Log
else:
    try:
        # Runtime imports
        from typing import runtime_checkable

        from typing_extensions import Protocol

        from .embedable import Embedable
        from .event import Event
        from .identifiable import Identifiable
        from .invokable import Invokable
        from .temporal import Temporal
        from .types import Embedding, Execution, ExecutionStatus, Log
    except ImportError:
        # Import error handling - define stub classes
        from ..utils.dependencies import check_protocols_dependencies

        def __getattr__(name):
            check_protocols_dependencies()
            raise ImportError(f"Cannot import {name} because dependencies are missing")


__all__ = [
    "Protocol",
    "runtime_checkable",
    "Identifiable",
    "Temporal",
    "Embedable",
    "Invokable",
    "Event",
    "Embedding",
    "ExecutionStatus",
    "Execution",
    "Log",
]



# Copyright (c) 2025, HaiyangLi <quantocean.li at gmail dot com>
#
# SPDX-License-Identifier: Apache-2.0

from enum import Enum

from pydantic import BaseModel, field_serializer, field_validator

from .utils import validate_model_to_dict

__all__ = ("Embedding", "ExecutionStatus", "Execution", "Metadata")


Embedding = list[float]
Metadata = dict


class ExecutionStatus(str, Enum):
    """Status states for tracking action execution progress."""

    PENDING = "pending"
    PROCESSING = "processing"
    COMPLETED = "completed"
    FAILED = "failed"


class Execution(BaseModel):
    """Represents the execution state of an event."""

    duration: float | None = None
    response: dict | None = None
    status: ExecutionStatus = ExecutionStatus.PENDING
    error: str | None = None

    @field_validator("response", mode="before")
    def _validate_response(cls, v: BaseModel | dict | None):
        return validate_model_to_dict(v)

    @field_serializer("status")
    def _serialize_status(self, v: ExecutionStatus) -> str:
        return v.value


class Log(BaseModel):
    id: str
    created_at: str
    updated_at: str
    event_type: str
    content: str | None = None
    embedding: Embedding = []
    duration: float | None = None
    status: str
    error: str | None = None
    sha256: str | None = None



# Copyright (c) 2025, HaiyangLi <quantocean.li at gmail dot com>
#
# SPDX-License-Identifier: Apache-2.0

from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator

from .utils import validate_uuid

__all__ = ("Identifiable",)


class Identifiable(BaseModel):
    """Base class for objects with a unique identifier and creation timestamp."""

    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
        arbitrary_types_allowed=True,
        populate_by_name=True,
    )

    id: UUID = Field(
        default_factory=uuid4,
        description="Unique identifier for the element.",
        frozen=True,
    )

    @field_serializer("id")
    def _serialize_ids(self, v: UUID) -> str:
        return str(v)

    @field_validator("id", mode="before")
    def _validate_ids(cls, v: str | UUID) -> UUID:
        return validate_uuid(v)

    def __hash__(self) -> int:
        """Returns the hash of the object."""
        return hash(self.id)



"""
Excel adapter (requires pandas + xlsxwriter engine).
"""

from __future__ import annotations

import io
from pathlib import Path
from typing import TypeVar

import pandas as pd
from pydantic import BaseModel

from ..core import Adapter
from ..exceptions import AdapterError, ResourceError
from .pandas_ import DataFrameAdapter

T = TypeVar("T", bound=BaseModel)


class ExcelAdapter(Adapter[T]):
    obj_key = "xlsx"

    # incoming
    @classmethod
    def from_obj(
        cls,
        subj_cls: type[T],
        obj: str | Path | bytes,
        /,
        *,
        many=True,
        sheet_name=0,
        **kw,
    ):
        try:
            if isinstance(obj, bytes):
                df = pd.read_excel(io.BytesIO(obj), sheet_name=sheet_name, **kw)
            else:
                df = pd.read_excel(obj, sheet_name=sheet_name, **kw)
            return DataFrameAdapter.from_obj(subj_cls, df, many=many)
        except FileNotFoundError as e:
            raise ResourceError(f"File not found: {e}", resource=str(obj)) from e
        except ValueError as e:
            raise AdapterError(
                f"Error adapting from xlsx (original_error='{e}')", adapter="xlsx"
            ) from e
        except Exception as e:
            raise AdapterError(
                f"Unexpected error in Excel adapter: {e}", adapter="xlsx"
            ) from e

    # outgoing
    @classmethod
    def to_obj(
        cls,
        subj: T | list[T],
        /,
        *,
        many=True,
        sheet_name="Sheet1",
        **kw,
    ) -> bytes:
        df = DataFrameAdapter.to_obj(subj, many=many)
        buf = io.BytesIO()
        with pd.ExcelWriter(buf, engine="xlsxwriter") as wr:
            df.to_excel(wr, sheet_name=sheet_name, index=False)
        return buf.getvalue()



# Copyright (c) 2025, HaiyangLi <quantocean.li at gmail dot com>
#
# SPDX-License-Identifier: Apache-2.0


import orjson as json
from pydantic import BaseModel, Field, field_validator

from .types import Embedding


class Embedable(BaseModel):
    """Embedable trait, contains embedding and content"""

    content: str | None = None
    embedding: Embedding = Field(default_factory=list)

    @property
    def n_dim(self) -> int:
        """Get the number of dimensions of the embedding."""
        return len(self.embedding)

    @field_validator("embedding", mode="before")
    def _parse_embedding(cls, value: list[float] | str | None) -> Embedding | None:
        if value is None:
            return []
        if isinstance(value, str):
            try:
                loaded = json.loads(value)
                return [float(x) for x in loaded]
            except Exception as e:
                raise ValueError("Invalid embedding string.") from e
        if isinstance(value, list):
            try:
                return [float(x) for x in value]
            except Exception as e:
                raise ValueError("Invalid embedding list.") from e
        raise ValueError("Invalid embedding type; must be list or JSON-encoded string.")

    def create_content(self):
        """override in child class to support custom content creation"""
        return self.content


def _parse_embedding_response(x):
    # parse openai response
    if (
        isinstance(x, BaseModel)
        and hasattr(x, "data")
        and len(x.data) > 0
        and hasattr(x.data[0], "embedding")
    ):
        return x.data[0].embedding

    if isinstance(x, list | tuple):
        if len(x) > 0 and all(isinstance(i, float) for i in x):
            return x
        if len(x) == 1 and isinstance(x[0], dict | BaseModel):
            return _parse_embedding_response(x[0])

    # parse dict response
    if isinstance(x, dict):
        # parse openai format response

        if "data" in x:
            data = x.get("data")
            if data is not None and len(data) > 0 and isinstance(data[0], dict):
                return _parse_embedding_response(data[0])

        # parse {"embedding": []} response
        if "embedding" in x:
            return _parse_embedding_response(x["embedding"])

    return x



"""
PostgresAdapter - thin preset over SQLAdapter (pgvector-ready if you add vec column).
"""

from __future__ import annotations

from typing import TypeVar

from pydantic import BaseModel

from ..exceptions import ConnectionError
from .sql_ import SQLAdapter

T = TypeVar("T", bound=BaseModel)


class PostgresAdapter(SQLAdapter[T]):  # type: ignore[type-arg]
    obj_key = "postgres"
    DEFAULT = "postgresql+psycopg://user:pass@localhost/db"

    @classmethod
    def from_obj(cls, subj_cls, obj: dict, /, **kw):
        try:
            # Set default connection string if not provided
            obj.setdefault("engine_url", cls.DEFAULT)

            # Add PostgreSQL-specific error handling
            try:
                return super().from_obj(subj_cls, obj, **kw)
            except Exception as e:
                # Check for common PostgreSQL-specific errors
                error_str = str(e).lower()
                if "authentication" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL authentication failed: {e}",
                        adapter="postgres",
                        url=obj["engine_url"],
                    ) from e
                elif "connection" in error_str and "refused" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL connection refused: {e}",
                        adapter="postgres",
                        url=obj["engine_url"],
                    ) from e
                elif "does not exist" in error_str and "database" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL database does not exist: {e}",
                        adapter="postgres",
                        url=obj["engine_url"],
                    ) from e
                # Re-raise the original exception
                raise

        except ConnectionError:
            # Re-raise ConnectionError
            raise
        except Exception as e:
            # Wrap other exceptions
            raise ConnectionError(
                f"Unexpected error in PostgreSQL adapter: {e}",
                adapter="postgres",
                url=obj.get("engine_url", cls.DEFAULT),
            ) from e

    @classmethod
    def to_obj(cls, subj, /, **kw):
        try:
            # Set default connection string if not provided
            kw.setdefault("engine_url", cls.DEFAULT)

            # Add PostgreSQL-specific error handling
            try:
                return super().to_obj(subj, **kw)
            except Exception as e:
                # Check for common PostgreSQL-specific errors
                error_str = str(e).lower()
                if "authentication" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL authentication failed: {e}",
                        adapter="postgres",
                        url=kw["engine_url"],
                    ) from e
                elif "connection" in error_str and "refused" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL connection refused: {e}",
                        adapter="postgres",
                        url=kw["engine_url"],
                    ) from e
                elif "does not exist" in error_str and "database" in error_str:
                    raise ConnectionError(
                        f"PostgreSQL database does not exist: {e}",
                        adapter="postgres",
                        url=kw["engine_url"],
                    ) from e
                # Re-raise the original exception
                raise

        except ConnectionError:
            # Re-raise ConnectionError
            raise
        except Exception as e:
            # Wrap other exceptions
            raise ConnectionError(
                f"Unexpected error in PostgreSQL adapter: {e}",
                adapter="postgres",
                url=kw.get("engine_url", cls.DEFAULT),
            ) from e



# Copyright (c) 2025, HaiyangLi <quantocean.li at gmail dot com>
#
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime, timezone

from pydantic import Field, field_serializer, field_validator

from .identifiable import Identifiable
from .utils import convert_to_datetime


class Temporal(Identifiable):
    """Allows for updating the last updated timestamp."""

    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
        description="Creation timestamp for the element.",
        frozen=True,
    )

    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
        description="Last updated timestamp for the element.",
    )

    def update_timestamp(self) -> None:
        """Update the last updated timestamp to the current time."""
        self.updated_at = datetime.now(timezone.utc)

    @field_serializer("updated_at", "created_at")
    def _serialize_datetime(self, v: datetime) -> str:
        return v.isoformat()

    @field_validator("updated_at", "created_at", mode="before")
    def _validate_datetime(cls, v: str | datetime) -> datetime:
        return convert_to_datetime(v)



"""
pydapter.core - Adapter protocol, registry, Adaptable mix-in.
"""

from __future__ import annotations

from typing import Any, ClassVar, Protocol, TypeVar, runtime_checkable

from .exceptions import (
    PYDAPTER_PYTHON_ERRORS,
    AdapterError,
    AdapterNotFoundError,
    ConfigurationError,
)

T = TypeVar("T", contravariant=True)


# ------------------------------------------------------------------ Adapter
@runtime_checkable
class Adapter(Protocol[T]):
    """Stateless conversion helper."""

    obj_key: ClassVar[str]

    @classmethod
    def from_obj(cls, subj_cls: type[T], obj: Any, /, *, many: bool = False, **kw): ...

    @classmethod
    def to_obj(cls, subj: T | list[T], /, *, many: bool = False, **kw): ...


# ----------------------------------------------------------- AdapterRegistry
class AdapterRegistry:
    def __init__(self) -> None:
        self._reg: dict[str, type[Adapter]] = {}

    def register(self, adapter_cls: type[Adapter]) -> None:
        key = getattr(adapter_cls, "obj_key", None)
        if not key:
            raise ConfigurationError(
                "Adapter must define 'obj_key'", adapter_cls=adapter_cls.__name__
            )
        self._reg[key] = adapter_cls

    def get(self, obj_key: str) -> type[Adapter]:
        try:
            return self._reg[obj_key]
        except KeyError as exc:
            raise AdapterNotFoundError(
                f"No adapter registered for '{obj_key}'", obj_key=obj_key
            ) from exc

    # convenience
    def adapt_from(self, subj_cls: type[T], obj, *, obj_key: str, **kw):
        try:
            result = self.get(obj_key).from_obj(subj_cls, obj, **kw)
            if result is None:
                raise AdapterError(f"Adapter {obj_key} returned None", adapter=obj_key)
            return result

        except Exception as exc:
            if isinstance(exc, AdapterError) or isinstance(exc, PYDAPTER_PYTHON_ERRORS):
                raise

            raise AdapterError(
                f"Error adapting from {obj_key}", original_error=str(exc)
            ) from exc

    def adapt_to(self, subj, *, obj_key: str, **kw):
        try:
            result = self.get(obj_key).to_obj(subj, **kw)
            if result is None:
                raise AdapterError(f"Adapter {obj_key} returned None", adapter=obj_key)
            return result

        except Exception as exc:
            if isinstance(exc, AdapterError) or isinstance(exc, PYDAPTER_PYTHON_ERRORS):
                raise

            raise AdapterError(
                f"Error adapting to {obj_key}", original_error=str(exc)
            ) from exc


# ----------------------------------------------------------------- Adaptable
class Adaptable:
    """Mixin that endows any Pydantic model with adapt-from / adapt-to."""

    _adapter_registry: ClassVar[AdapterRegistry | None] = None

    # registry
    @classmethod
    def _registry(cls) -> AdapterRegistry:
        if cls._adapter_registry is None:
            cls._adapter_registry = AdapterRegistry()
        return cls._adapter_registry

    @classmethod
    def register_adapter(cls, adapter_cls: type[Adapter]) -> None:
        cls._registry().register(adapter_cls)

    # high-level helpers
    @classmethod
    def adapt_from(cls, obj, *, obj_key: str, **kw):
        return cls._registry().adapt_from(cls, obj, obj_key=obj_key, **kw)

    def adapt_to(self, *, obj_key: str, **kw):
        return self._registry().adapt_to(self, obj_key=obj_key, **kw)
